#!/usr/bin/env python3

import amulet
import sys
import time
import traceback
import requests
import pymongo
from pymongo import MongoClient
from collections import Counter


#########################################################
# Test Quick Config
#########################################################
scale = 3
seconds = 1800

# max amount of time to wait before testing for replicaset
# status
wait_for_replicaset = 600

#########################################################
# 3shard cluster configuration
#########################################################
d = amulet.Deployment(series='trusty')

d.add('mongodb', charm='mongodb', units=scale)
d.expose('mongodb')

# Perform the setup for the deployment.
try:
    d.setup(seconds)
    d.sentry.wait(seconds)
except amulet.helpers.TimeoutError:
    message = 'The environment did not setup in %d seconds.', seconds
    amulet.raise_status(amulet.SKIP, msg=message)
except:
    raise

sentry_dict = {
    'mongodb0-sentry': d.sentry['mongodb'][0],
    'mongodb1-sentry': d.sentry['mongodb'][1],
    'mongodb2-sentry': d.sentry['mongodb'][2],
}


#############################################################
# Check presence of MongoDB GUI HEALTH Status
#############################################################
def validate_status_interface():
    pubaddy = d.sentry['mongodb'][0].info['public-address']
    fmt = "http://{}:28017"
    if ":" in pubaddy:
        fmt = "http://[{}]:28017"
    r = requests.get(fmt.format(pubaddy), verify=False)
    r.raise_for_status


#############################################################
# Validate that each unit has an active mongo service
#############################################################
def validate_running_services():
    for service in sentry_dict:
        output = sentry_dict[service].run('service mongodb status')
        service_active = str(output).find('mongodb start/running')
        if service_active == -1:
            message = "Failed to find running MongoDB on host {}".format(
                service)
            amulet.raise_status(amulet.SKIP, msg=message)


#############################################################
# Validate proper replicaset setup
#############################################################
def validate_replicaset_setup():

    d.sentry.wait(seconds)

    unit_status = []
    time_between = 10
    tries = wait_for_replicaset / time_between

    for service in sentry_dict:
        addy = sentry_dict[service].info['public-address']
        if ":" in addy:
            addy = "[{}]".format(addy)
        while True:
            try:
                client = MongoClient(addy)
                r = client.admin.command('replSetGetStatus')
                break
            except pymongo.errors.OperationFailure as ex:
                sys.stderr.write(
                    'OperationFailure, sleep and retry... to {}: {}\n'.
                    format(addy, ex))
                tb_lines = traceback.format_exception(ex.__class__,
                                                      ex, ex.__traceback__)
                tb_text = ''.join(tb_lines)
                sys.stderr.write(tb_text)
                tries = tries - 1
                if tries < 0:
                    sys.stderr.write('retry limit caught, failing...\n')
                    break
                time.sleep(time_between)
        unit_status.append(r['myState'])
        client.close()

    primaries = Counter(unit_status)[1]
    if primaries != 1:
        message = "Only one PRIMARY unit allowed! Found: %s %s" % (primaries,
                                                                   unit_status)
        amulet.raise_status(amulet.FAIL, message)

    secondrs = Counter(unit_status)[2]
    if secondrs != 2:
        message = ("Only two SECONDARY units allowed! (Found %s) %s" %
                   (secondrs, unit_status))
        amulet.raise_status(amulet.FAIL, message)


#############################################################
# Validate connectivity from $WORLD
#############################################################
def validate_world_connectivity():
    d.sentry['mongodb']
    ordered_units = sorted(d.sentry['mongodb'], key=lambda u: u.info['unit'])
    # We assume minimum unit number is master.
    addy = ordered_units[0].info['public-address']
    if ":" in addy:
        addy = "[{}]".format(addy)

    time_between = 10
    tries = wait_for_replicaset / time_between
    insert_id = None
    while True:
        try:
            client = MongoClient(addy)
            db = client.test
            # Can we successfully insert?
            insert_id = db.amulet.insert({'assert': True})
            break
        except pymongo.errors.AutoReconnect as ex:
            sys.stderr.write(
                'AutoReconnect error, sleep and retry... to {}: {}\n'.
                format(addy, ex))
            tb_lines = traceback.format_exception(ex.__class__,
                                                  ex, ex.__traceback__)
            tb_text = ''.join(tb_lines)
            sys.stderr.write(tb_text)
            tries = tries - 1
            if tries < 0:
                sys.stderr.write('retry limit caught, failing...\n')
                break
            time.sleep(time_between)

    if insert_id is None:
        amulet.raise_status(amulet.FAIL, msg="Failed to insert test data")
    # Can we delete from a shard using the Mongos hub?
    result = db.amulet.remove(insert_id)
    if result['err'] is not None:
        amulet.raise_status(amulet.FAIL, msg="Failed to remove test data")


validate_status_interface()
validate_running_services()
validate_replicaset_setup()
validate_world_connectivity()
